from typing import List, Set
from enum import Enum
from sand.config import SETTINGS
from sand.models.ontology import OntProperty, OntPropertyAR, OntPropertyDataType
from sand.models.table import Table, TableRow
import sm.outputs.semantic_model as O
from sm.misc.funcs import import_attr
from hugedict.chained_mapping import ChainedMapping
from drepr.models import (
    DRepr,
    Attr,
    Resource,
    ResourceType,
    CSVProp,
    Path,
    IndexExpr,
    RangeExpr,
    RangeAlignment,
    AlignedStep,
    Preprocessing,
    PreprocessingType,
    PMap,
)
from drepr.engine import execute, OutputFormat, MemoryOutput
from slugify import slugify
from sand.extension_interface.export import IExport
from sand.extensions.export.drepr.resources import get_entity_resource, get_table_resource
from sand.extensions.export.drepr.semanticmodel import get_drepr_sm, get_entity_data_nodes
from sand.extensions.export.drepr.transformation import has_transformation, get_transformation


class DreprExport(IExport):
    def export_data_model(self, table: Table, sm: O.SemanticModel) -> DRepr:
        """Create a D-REPR model of the dataset."""
        columns = [slugify(c).replace("-", "_") for c in table.columns]
        get_attr_id = lambda ci: f"{ci}__{columns[ci]}"
        get_ent_attr_id = lambda ci: f"{ci}__ent__{columns[ci]}"
        ent_dnodes = get_entity_data_nodes(sm)

        attrs = [
            Attr(
                id=get_attr_id(ci),
                resource_id="table",
                path=Path(
                    steps=[
                        RangeExpr(start=1, end=table.size + 1, step=1),
                        IndexExpr(val=ci),
                    ]
                ),
                missing_values=[""],
            )
            for ci in range(len(table.columns))
        ]
        attrs += [
            Attr(
                id=get_ent_attr_id(node.col_index),
                resource_id="entity",
                path=Path(
                    steps=[
                        RangeExpr(start=1, end=table.size + 1, step=1),
                        IndexExpr(val=node.col_index),
                    ]
                ),
                missing_values=[""],
            )
            for node in ent_dnodes
        ]

        id2props = ChainedMapping(
            OntPropertyAR(),
            import_attr(SETTINGS["ont_props"]["default"]),
        )
        dsm = get_drepr_sm(sm, id2props, get_attr_id, get_ent_attr_id)

        datatype_transformations = []
        for node in sm.nodes():
            if not isinstance(node, O.DataNode):
                continue

            datatypes: Set[OntPropertyDataType] = {
                id2props[OntProperty.uri2id(inedge.abs_uri)].datatype
                for inedge in sm.in_edges(node.id)
            }
            datatype = list(datatypes)[0] if len(datatypes) == 1 else None
            if datatype is None or not has_transformation(datatype):
                continue
            datatype_transformations.append(
                Preprocessing(
                    type=PreprocessingType.pmap,
                    value=PMap(
                        resource_id="table",
                        path=Path(
                            steps=[
                                RangeExpr(start=1, end=table.size + 1, step=1),
                                IndexExpr(val=node.col_index),
                            ]
                        ),
                        code=get_transformation(datatype),
                        change_structure=False,
                    ),
                )
            )

        return DRepr(
            resources=[
                Resource(id="table", type=ResourceType.CSV, prop=CSVProp()),
                Resource(id="entity", type=ResourceType.CSV, prop=CSVProp()),
            ],
            preprocessing=datatype_transformations,
            attrs=attrs,
            aligns=[
                       RangeAlignment(
                           source=get_attr_id(0),
                           target=get_attr_id(ci),
                           aligned_steps=[AlignedStep(source_idx=0, target_idx=0)],
                       )
                       for ci in range(1, len(table.columns))
                   ]
                   + [
                       RangeAlignment(
                           source=get_attr_id(0),
                           target=get_ent_attr_id(node.col_index),
                           aligned_steps=[AlignedStep(source_idx=0, target_idx=0)],
                       )
                       for node in ent_dnodes
                   ],
            sm=dsm,
        )

    def export_data(self, table: Table, rows: List[TableRow], sm: O.SemanticModel,
                    output_format: OutputFormat):
        """Convert a relational table into RDF format"""
        if len(table.columns) == 0:
            # no column, no data
            return ""

        ent_columns = {node.col_index for node in get_entity_data_nodes(sm)}
        resources = {
            "table": get_table_resource(table, rows),
            "entity": get_entity_resource(table, rows, ent_columns),
        }

        content = execute(
            ds_model=self.export_data_model(table, sm),
            resources=resources,
            output=MemoryOutput(output_format),
            debug=False,
        )
        return content



import csv
from dataclasses import dataclass
from io import StringIO
from typing import Dict, List, Set, Tuple
from uuid import uuid4

import orjson
from drepr.models import ResourceData, ResourceDataString

from sand.config import SETTINGS
from sand.models.entity import NIL_ENTITY_ID, Entity
from sand.models.ontology import OntPropertyAR
from sand.models.table import Table, TableRow


def get_table_resource(table: Table, rows: List[TableRow]) -> ResourceData:
    """Return a CSV resource, in which the first row is the header"""
    f = StringIO()
    writer = csv.writer(
        f, delimiter=",", quoting=csv.QUOTE_MINIMAL, lineterminator="\n"
    )

    writer.writerow(table.columns)
    for row in rows:
        writer.writerow(row.row)

    return ResourceDataString(f.getvalue())


def get_entity_resource(
    table: Table, rows: List[TableRow], ent_columns: Set[int]
) -> ResourceData:
    """Return a CSV resource of matrix mapping each position in a table to a corresponding entity uri."""
    new_entity_template: str = SETTINGS["entity"]["new_entity_template"]
    ent_rows: List[List[str]] = []

    for ri, row in enumerate(rows):
        ent_rows.append([])
        for ci in range(len(row.row)):
            if ci not in ent_columns:
                ent_rows[-1].append("")
                continue

            links = row.links.get(str(ci), [])
            # TODO: what happens when there are multiple entities?
            # for now, just return the first entity
            ent = None
            for link in links:
                if link.entity_id is not None and link.entity_id != NIL_ENTITY_ID:
                    ent = Entity.id2uri(link.entity_id)
                    break
            else:
                # generate new entity
                ent = new_entity_template.format(id=str(uuid4()))

            ent_rows[-1].append(ent)

    f = StringIO()
    writer = csv.writer(
        f, delimiter=",", quoting=csv.QUOTE_MINIMAL, lineterminator="\n"
    )

    writer.writerow(table.columns)
    for row in ent_rows:
        writer.writerow(row)
    return ResourceDataString(f.getvalue())



from dataclasses import dataclass
from itertools import chain
from typing import Mapping

from hugedict.misc import identity
from kgdata.wikidata import db
from kgdata.wikidata.models import WDClass, WDEntity, WDProperty, WDValue
from sm.namespaces.wikidata import WikidataNamespace

from sand.models.base import StoreWrapper
from sand.models.entity import DEFAULT_ENTITY, Entity, Statement, Value
from sand.models.ontology import (
    DEFAULT_ONT_CLASSES,
    DEFAULT_ONT_PROPS,
    OntClass,
    OntProperty,
    OntPropertyDataType,
)

wdns = WikidataNamespace.create()
WD_ONT_CLASSES = {
    wdns.get_rel_uri(wdns.STATEMENT_URI): OntClass(
        id=wdns.get_rel_uri(wdns.STATEMENT_URI),
        uri=wdns.STATEMENT_URI,
        label=wdns.get_rel_uri(wdns.STATEMENT_URI),
        aliases=[],
        description="Describes the claim of a statement and list references for this claim",
        parents=[],
    )
}
WD_ONT_CLASSES.update(DEFAULT_ONT_CLASSES)
INVERSE_DEFAULT_URI2ID = {
    v.uri: k
    for k, v in chain(
        DEFAULT_ONT_PROPS.items(), WD_ONT_CLASSES.items(), DEFAULT_ENTITY.items()
    )
}
DEFAULT_ID2URI = {
    k: v.uri
    for k, v in chain(
        DEFAULT_ONT_PROPS.items(), WD_ONT_CLASSES.items(), DEFAULT_ENTITY.items()
    )
}

WD_DATATYPE_MAPPING: Mapping[str, OntPropertyDataType] = {
    "monolingualtext": "monolingualtext",
    "url": "url",
    "wikibase-item": "entity",
    "external-id": "entity",
    "time": "datetime",
    "quantity": "number",
    "string": "string",
    "globe-coordinate": "globe-coordinate",
}


@dataclass
class WrapperWDEntity(Entity):
    @property
    def readable_label(self):
        return f"{self.label} ({self.id})"

    @property
    def instanceof(self):
        """Get the property representing the instanceof relation."""
        return "P31"


@dataclass
class WrapperWDClass(OntClass):
    @property
    def readable_label(self):
        return f"{self.label} ({self.id})"


@dataclass
class WrapperWDProperty(OntProperty):
    @property
    def readable_label(self):
        return f"{self.label} ({self.id})"


def get_entity_db(dbfile: str, proxy: bool):
    store = db.get_entity_db(dbfile, proxy=proxy, read_only=not proxy)
    return StoreWrapper(
        store,
        key_deser=identity,
        val_deser=qnode_deser,
    )


def get_ontclass_db(dbfile: str, proxy: bool):
    return StoreWrapper(
        db.get_wdclass_db(dbfile, proxy=proxy, read_only=not proxy),
        key_deser=get_wdclass_id,
        val_deser=ont_class_deser,
    )


def get_ontprop_db(dbfile: str, proxy: bool):
    return StoreWrapper(
        db.get_wdprop_db(dbfile, proxy=proxy, read_only=not proxy),
        key_deser=get_wdprop_id,
        val_deser=ont_prop_deser,
    )


def qnode_deser(qnode: WDEntity):
    props = {}
    for pid, stmts in qnode.props.items():
        new_stmts = []
        for stmt in stmts:
            new_stmt = Statement(
                value=wd_value_deser(stmt.value),
                qualifiers={
                    qid: [wd_value_deser(x) for x in lst]
                    for qid, lst in stmt.qualifiers.items()
                },
                qualifiers_order=stmt.qualifiers_order,
            )
            new_stmts.append(new_stmt)
        props[pid] = new_stmts

    return WrapperWDEntity(
        id=qnode.id,
        uri=wdns.get_entity_abs_uri(qnode.id),
        label=qnode.label,
        aliases=qnode.aliases,
        description=qnode.description,
        properties=props,
    )


def ont_class_deser(item: WDClass):
    return WrapperWDClass(
        id=item.id,
        uri=WikidataNamespace.get_entity_abs_uri(item.id),
        aliases=item.aliases,
        label=item.label,
        description=item.description,
        parents=item.parents,
        ancestors=item.ancestors,
    )


def ont_prop_deser(item: WDProperty):
    global WD_DATATYPE_MAPPING
    return WrapperWDProperty(
        id=item.id,
        uri=WikidataNamespace.get_prop_abs_uri(item.id),
        aliases=item.aliases,
        label=item.label,
        description=item.description,
        datatype=WD_DATATYPE_MAPPING[item.datatype]
        if item.datatype in WD_DATATYPE_MAPPING
        else "unknown",
        parents=item.parents,
        ancestors=item.ancestors,
    )


def wd_value_deser(val: WDValue):
    if val.is_entity_id(val):
        return Value(type="entityid", value=val.as_entity_id())
    elif val.is_string(val):
        return Value(type="string", value=val.as_string())
    else:
        return Value(type=val.type, value=val.value)


def get_wdprop_id(uri_or_id: str):
    if uri_or_id.startswith("http"):
        return uri_or_id.replace(f"http://www.wikidata.org/prop/", "")
    return uri_or_id


def get_wdclass_id(uri_or_id: str):
    if uri_or_id.startswith("http"):
        return uri_or_id.replace(f"http://www.wikidata.org/entity/", "")
    return uri_or_id


def uri2id(uri: str):
    if uri.startswith("http://www.wikidata.org/prop/"):
        return WikidataNamespace.get_prop_id(uri)
    if uri.startswith("http://www.wikidata.org/entity/"):
        return WikidataNamespace.get_entity_id(uri)
    if uri in INVERSE_DEFAULT_URI2ID:
        return INVERSE_DEFAULT_URI2ID[uri]
    return uri


def id2uri(id: str):
    if id in DEFAULT_ID2URI:
        return DEFAULT_ID2URI[id]
    if id.startswith("P"):
        return wdns.get_prop_abs_uri(id)
    if id.startswith("Q"):
        return wdns.get_entity_abs_uri(id)
    raise ValueError(f"Cannot convert unknown id to uri: {id}")


def get_rel_uri(uri: str):
    return wdns.get_rel_uri(uri)



import json
import threading
from typing import Dict, List, Union, Literal
from flask.blueprints import Blueprint
from sm.misc.funcs import import_func
from sand.config import SETTINGS
from flask import request, jsonify, abort

from sand.extension_interface.search import IEntitySearch, IOntologySearch
from sand.models.search import SearchResult
from gena.serializer import get_dataclass_serializer

search_bp = Blueprint("search", "search")

GetSearchCache = threading.local()
serializer = get_dataclass_serializer(SearchResult)


def get_search(name: Literal['classes', 'entities', 'props']) -> Union[IEntitySearch, IOntologySearch]:
    """
    Returns an implementation of an ISearch Interface from the
    configuration file.
    """
    global GetSearchCache

    if not hasattr(GetSearchCache, "search"):
        GetSearchCache.search = {}
    search_config = SETTINGS["search"]
    constructor = search_config[name]
    GetSearchCache.search[name] = import_func(constructor)()

    return GetSearchCache.search[name]


@search_bp.route(f"/{search_bp.name}/classes", methods=["GET"])
def search_classes():
    """API Route to search for classes with their names"""
    search_text = request.args.get('q')
    wikidata_search = get_search('classes')
    search_results = wikidata_search.find_class_by_name(search_text)
    serialized_payload = [serializer(item) for item in search_results]
    return jsonify({'items': serialized_payload})


@search_bp.route(f"/{search_bp.name}/entities", methods=["GET"])
def search_entities():
    """API Route to search for entities with their names"""
    search_text = request.args.get('q')
    wikidata_search = get_search('entities')
    search_results = wikidata_search.find_entity_by_name(search_text)
    serialized_payload = [serializer(item) for item in search_results]
    return jsonify({'items': serialized_payload})


@search_bp.route(f"/{search_bp.name}/props", methods=["GET"])
def search_props():
    """API Route to search for properties with their names"""
    search_text = request.args.get('q')
    wikidata_search = get_search('props')
    search_results = wikidata_search.find_props_by_name(search_text)
    serialized_payload = [serializer(item) for item in search_results]
    return jsonify({'items': serialized_payload})



from pathlib import Path
from typing import List, Tuple

import click
from sm.dataset import Dataset, Example, FullTable
from tqdm.auto import tqdm

from sand.models import (
    ContextPage,
    Link,
    Project,
    SemanticModel,
    Table,
    TableRow,
    Value,
)
from sand.models import db as dbconn
from sand.models import init_db


@click.command(name="load")
@click.option("-d", "--db", required=True, help="smc database file")
@click.option("-p", "--project", default="default", help="Project name")
@click.option("--dataset", required=True, help="Path to tables")
@click.option(
    "-n",
    "--n-tables",
    default=-1,
    type=int,
    help="Number of tables to load (negative number or zero to load all)",
)
def load_dataset(db: str, project: str, dataset: str, n_tables: int):
    """Load a dataset into a project"""
    init_db(db)

    examples = Dataset(Path(dataset)).load()

    if n_tables > 0:
        examples = examples[:n_tables]

    with dbconn:
        p = Project.get(name=project)
        for e in tqdm(examples, desc="Loading examples"):
            save_example(p, e)


def save_example(project: Project, example: Example[FullTable]):
    mtbl, mrows = convert_linked_table(example.table)
    mtbl.project = project
    mtbl.save()

    for row in mrows:
        row.save()

    for i, sm in enumerate(example.sms):
        SemanticModel(
            table=mtbl,
            name=f"sm-{i}",
            description="",
            version=1,
            data=sm,
        ).save()


def convert_linked_table(tbl: FullTable) -> Tuple[Table, List[TableRow]]:
    """Convert linked table to smc's tables and rows.
    Note that the project of the table is not set, which is supposed to be set later.
    """
    context_values = [
        Value("entityid", str(entityid)) for entityid in tbl.context.page_entities
    ]

    if tbl.context.page_url is not None:
        context_page = ContextPage(
            tbl.context.page_url,
            tbl.context.page_title or "",
            str(tbl.context.page_entities[0])
            if len(tbl.context.page_entities) > 0
            else None,
        )
    else:
        context_page = None

    nrows, ncols = tbl.table.shape()
    links = {}
    for ri in range(nrows):
        links[ri] = {}

    for ri, ci, lst in tbl.links.enumerate_flat_iter():
        if len(lst) == 0:
            continue
        links[ri][ci] = [
            Link(
                start=l.start,
                end=l.end,
                url=l.url,
                entity_id=str(l.entities[0]) if len(l.entities) > 0 else None,
                candidate_entities=[],  # do not load the entities because this is filled by methods
            )
            for l in lst
        ]

    columns = [col.name for col in tbl.table.columns]
    mtbl = Table(
        name=tbl.table.table_id,
        description="",
        columns=columns,
        size=nrows,
        context_values=context_values,
        context_tree=[],
        context_page=context_page,
        project=None,
    )
    mrows = []

    for ri in range(nrows):
        row = [tbl.table[ri, ci] for ci in range(len(columns))]
        mrows.append(TableRow(table=mtbl, index=ri, row=row, links=links[ri]))

    return mtbl, mrows



from typing import List
import re
from sand.extension_interface.search import IEntitySearch, IOntologySearch
from sm.misc.funcs import import_attr
from sand.models.search import SearchResult
from sand.config import SETTINGS


class DefaultSearch(IEntitySearch, IOntologySearch):

    def __init__(self):
        self.default_classes = import_attr(SETTINGS['ont_classes']['default'])
        self.default_entities = import_attr(SETTINGS['entity']['default'])
        self.default_properties = import_attr(SETTINGS['ont_props']['default'])

    def local_search(self, default_entities: dict, search_text: str) -> List:
        """ performs local partial text search across default entities"""
        query_tokens = re.findall(r'[a-z]+|\d+', search_text.lower())
        search_results = []

        if not query_tokens:
            return search_results

        for entity in default_entities.values():
            label = entity.label.lower()
            if all(token in label for token in query_tokens):
                search_results.append(entity)
        return search_results

    def find_class_by_name(self, search_text: str) -> List[SearchResult]:
        """Uses local default classes to search for entities using their name/text."""
        local_search_results = self.local_search(self.default_classes, search_text)
        payload_results = []

        for entity in local_search_results:
            item = SearchResult(
                label=entity.label,
                id=entity.id,
                description=entity.description,
                uri=entity.uri
            )
            payload_results.append(item)

        return payload_results

    def find_entity_by_name(self, search_text: str) -> List[SearchResult]:
        """Uses local default entities to search for entities using their name/text."""
        local_search_results = self.local_search(self.default_entities, search_text)
        payload_results = []

        for entity in local_search_results:
            item = SearchResult(
                label=entity.label,
                id=entity.id,
                description=entity.description,
                uri=entity.uri
            )
            payload_results.append(item)

        return payload_results

    def find_props_by_name(self, search_text: str) -> List[SearchResult]:
        """Uses local default properties to search for entities using their name/text."""
        local_search_results = self.local_search(self.default_properties, search_text)
        payload_results = []

        for entity in local_search_results:
            item = SearchResult(
                label=entity.label,
                id=entity.id,
                description=entity.description,
                uri=entity.uri
            )
            payload_results.append(item)

        return payload_results



from abc import ABC, abstractmethod
from typing import List

import sm.outputs.semantic_model as O
from sand.models.table import Table, TableRow
from drepr.models import DRepr
from drepr.engine import OutputFormat


class IExport(ABC):
    """
    Export interface class to export relational data to
    different data formats
    """

    @abstractmethod
    def export_data_model(self, table: Table, sm: O.SemanticModel) -> DRepr:
        """export data model using DREPR"""
        pass

    @abstractmethod
    def export_data(self, table: Table, rows: List[TableRow], sm: O.SemanticModel,
                    output_format: OutputFormat):
        """export relational data using DREPR"""
        pass
